﻿import os
import binascii
import ConfigParser
import glob
import sys

def read2(file, verbal=False):
    data = os.read(file, 2)
    bit1 = ord(data[0])
    bit2 = ord(data[1])
    sum = bit2 * 256 + bit1
    if verbal:
        print "BIT1: ", bit1
        print "BIT2: ", bit2
        print "SUM : ", sum
    return sum

def sread(file, num, conf):
    # Read bytes
    array = []
    temp = ""
    for x in range(0,num):
        array.append(binascii.hexlify(str(os.read(file, 1))))
    for x in range(0,num):
        try:
            temp += chr(int(array[x], 16))
        except ValueError:
            temp += "?"
    temp = temp[:-1]
    return temp

def parse_oop(raw, conf):
    # Parse data
    ret = ""
    raw = raw.split("\r")
    last_text = False
    for line in raw:
        if line == "":
            line = conf["ending"]
        if conf["dump"] == "play" and line[:5] == "#play":
            ret += line + conf["ending"]
        
        if conf["dump"] == "text" and line[0] not in ["@", "#", ":", "/", "?", "!", "'"]:
            if line[0] == "$":
                line = line[1:]
            ret += line + conf["ending"]
            
        
        if conf["dump"] == "oop":
            ret += line + conf["ending"]
    return ret[:-1*(len(conf["ending"]))]

def main():
    if os.name == "nt":
        os.system("cls")
    else:
        os.system("clear")
    # File flags
    flags = os.O_RDONLY
    if hasattr(os, 'O_BINARY'):
        flags = flags | os.O_BINARY
    
    # Load config
    print "ZZT-POOP by Dr. Dos"
    print "="*60
    
    config = ConfigParser.RawConfigParser()
    config.read("zzt_poop.ini")
    
    
    conf = {}
    conf["dump"]        = config.get("Settings", "dump")
    conf["output"]      = config.get("Settings", "output")
    conf["ending"]      = "\r"+(os.name == "nt")*"\n"
    conf["obj_delim"]   = config.get("Delimiters", "object")
    conf["room_delim"]  = config.get("Delimiters", "room")
    conf["world_delim"] = config.get("Delimiters", "world")
    conf["coords"]      = config.getboolean("Extra", "coords")
    conf["names"]      = config.getboolean("Extra", "names")
    conf["titles"]      = config.getboolean("Extra", "titles")
    conf["worlds"]      = config.getboolean("Extra", "worlds")
    
    if conf["obj_delim"] == "NONE":
        conf["obj_delim"] = ""
    if conf["room_delim"] == "NONE":
        conf["room_delim"] = ""
    if conf["world_delim"] == "NONE":
        conf["world_delim"] = ""
    
    print "CURRENT SETTINGS"
    if conf["dump"] == "text":
        print "Dump              : Text Only"
    elif conf["dump"] == "play":
        print "Dump              : #play Only"
    elif conf["dump"] == "oop":
        print "Dump              : All ZZT-OOP"
    else:
        print "ERROR: Invalid dump settings!"
        
    if conf["output"] == "split":
        print "Output            : One file per world."
    elif conf["dump"] == "combined":
        print "Output            : Combined text file"
    else:
        print "ERROR: Invalid output settings!"
    
    if os.name == "nt":
        print "Line Endings      : Windows (CRLF)"
    else:
        print "Line Endings      : Linux (LF)"
    
    print "World  Delimiter  : " + conf["world_delim"]*7
    print "Room   Delimiter  : " + conf["room_delim"]*6
    print "Object Delimiter  : " + conf["obj_delim"]*5
    
    print "Show Object Names : " + "Yes"*(conf["names"] == 1) + "No"*(conf["names"]==0)
    print "Show Coordinates  : " + "Yes"*(conf["coords"] == 1) + "No"*(conf["coords"]==0)
    print "Show Room Names   : " + "Yes"*(conf["titles"] == 1) + "No"*(conf["titles"]==0)
    print "Show Filenames    : " + "Yes"*(conf["worlds"] == 1) + "No"*(conf["worlds"]==0)
    
    print "="*60
    raw_input("Press Enter to begin ")
    
    # Load files
    if len(sys.argv) > 1:
        print "Loading your world list..."
        worlds = sys.argv[1:]
    else:
        print "Scanning for ZZT worlds..."
        worlds = glob.glob("*.[Zz][Zz][Tt]")
    
    if len(worlds) < 1:
        print "ERROR: No ZZT worlds found!"
        sys.exit()
    
    world_count = 1
    combined_output = ""
    for fname in worlds:
        file = os.open(fname, flags)
        print os.path.basename(fname)
    
        # World Header
        world_header = conf["world_delim"]*80 + conf["ending"]  * (conf["world_delim"] != "")
        if conf["worlds"]:
            world_header += conf["world_delim"] * 2 + " WORLD  : " + os.path.basename(fname) + conf["ending"]
            world_header += conf["world_delim"]*80 + conf["ending"]
        world_output = ""
        # Skip ZZT file flag
        os.lseek(file, 2, 1)
        
        # Get number of boards
        boardNum = read2(file) + 1 # Plus one for title screen
    
        # Skip the rest of the world header (508 bytes)
        os.read(file, 508)
        
        cur_board = 0
        
        while (cur_board != boardNum):
            
            #Tile we're looking at
            x = 1
            y = 1
            
            # Get board size
            boardSize = read2(file)
            title_length = int(hex(ord(os.read(file, 1))), 16)
            
            # Title
            title = os.read(file, 50)[:title_length]
            print "\t", title, "..."
            
            # Board header
            board_output = ""
            board_header = conf["room_delim"]*70 + conf["ending"] * (conf["room_delim"] != "")
            if conf["titles"]:
                board_header += conf["room_delim"] * 2 + " ROOM   : " + title + conf["ending"]
                board_header += conf["room_delim"]*70 + conf["ending"]
            
            # There are now 1500 tiles to be declared, find all the scrolls and objects
            remaining = 1500
            important = []
            while remaining:
                tiles = os.read(file, 3)
                
                try:
                    chunk = int(hex(ord(tiles[0])), 16)
                    
                    type = int(hex(ord(tiles[1])), 16)
                    if type in [10, 36]: # Scroll and Object IDs
                        important.append((x,y,type))
                except:
                    break
                remaining -= chunk
                x += chunk
                while x > 60:
                    x -= 60
                    y += 1
            
            os.read(file, 70) # Various board properties
            
            os.read(file, 16) # Unknown
            
            stats = read2(file) + 1 # Number of stat elements including player
            # Read stat elements
            
            has_code = False
            for count in xrange(0, stats):
                x = int(hex(ord(os.read(file, 1))), 16)
                y = int(hex(ord(os.read(file, 1))), 16)
                #print x, "/", y
                
                os.read(file, 21) # Various stat properties
                
                oop_size = read2(file)
                if oop_size < 0 or oop_size > 32767:
                    oop_size = 0
                    
                os.read(file, 8) # Padding
                
                # Object Header
                obj_header = conf["obj_delim"]*60 + conf["ending"] * (conf["obj_delim"] != "")
                if conf["names"]:
                    obj_header += conf["obj_delim"] * 2 + " " + "OBJECT_NAME" + conf["ending"]
                if conf["coords"]:
                    obj_header += conf["obj_delim"] * 2 + " " + "COORDS : ["+str(x)+", "+str(y)+"]" + conf["ending"]
                if conf["names"] or conf["coords"]:
                    obj_header += conf["obj_delim"]*60 + conf["ending"]
                
                my_code = ""
                if oop_size > 0: # ZZT-OOP
                    has_code = True
                    my_code = sread(file, oop_size, conf)
                    my_code = parse_oop(my_code, conf)
                
                    if conf["names"]:
                        if my_code.find("@") == 0:
                            obj_header = obj_header.replace("OBJECT_NAME", "NAME   : "+my_code[1:my_code.find(conf["ending"])])
                        elif type == 10:
                            obj_header = obj_header.replace("OBJECT_NAME", "NAME   : Scroll")
                        else:
                            obj_header = obj_header.replace("OBJECT_NAME", "NAME   : Interaction")
                
                    board_output += obj_header + my_code + conf["ending"] * 2
            
            if has_code:
                board_output = board_header + board_output
            world_output += board_output
            cur_board += 1
         
        # Write individual files if files are split
        if conf["output"] == "split":
            output = open(os.path.basename(fname)+"-"+conf["dump"]+".txt", "w")
            output.write(world_header + world_output)
            output.close()
            print "Wrote "+os.path.basename(fname)[:-4]+"-"+conf["dump"]+".txt"
        else:
            combined_output += world_header + world_output
            
        # Move to next world
        world_count += 1
        if world_count <= len(worlds):
            combined_output += conf["ending"]*4
    
    # Write combined file if files are combined
    if conf["output"] == "combined":
        output = open("combined-"+conf["dump"]+".txt", "w")
        output.write(combined_output)
        output.close()
        print "Wrote combined-"+conf["dump"]+".txt"
        
    print "DONE."
    sys.exit()
    
if __name__ == "__main__": main()   